import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';

import 'observable_model.dart';
import 'service.dart';

/// A function used to initialize values in the [ObservableModel].
///
/// It is executed once when [AppGadget] initializes its state.
typedef Init = void Function(ObservableModel model);

/// A function called when the state of [AppGadget] is disposed.
typedef Dispose = void Function(BuildContext context);

/// A Gadget used as the root for a Flutter Gadgets application.
///
/// Initializes and makes available to its descendat the given [Service]s
/// and creates a new [ObservableModel] instance.
///
/// An optional [Init] function can be provided in order to initialize
/// some values in the [ObservableModel].
///
/// It is used as follows:
///
/// ```dart
/// void main() => runApp(AppGadget(
///   services: [
///     Service(create: (_) => MyDAO()),
///     Service(create: (_) => MyOtherDAO()),
///     Service(create: (_) => MyExample()),
///   ], // optional
///   init: (model) => model.putValue<MyValue>(MyValue()) // optional
///   dispose: (context) { //Do some stuff... } // optional
///   child: MyChild
/// ));
/// ```
class AppGadget extends StatefulWidget {
  const AppGadget({
    Key? key,
    required this.child,
    this.init,
    this.dispose,
    this.services,
  }) : super(key: key);

  /// The widget tree wrapped by [AppGadget].
  final Widget child;
  final Init? init;
  final Dispose? dispose;

  /// The list of [Service]s available to the [AppGadget] descendant.
  final List<Service>? services;

  @override
  _AppGadgetState createState() => _AppGadgetState();
}

class _AppGadgetState extends State<AppGadget> {
  final _model = ObservableModel();

  @override
  void initState() {
    super.initState();
    if (widget.init != null) {
      widget.init!(_model);
    }
  }

  @override
  void dispose() {
    if (widget.dispose != null) {
      widget.dispose!(context);
    }
    _model.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return MultiProvider(
      providers: [
        Provider<ObservableModel>.value(value: _model),
        ...services,
      ],
      child: widget.child,
    );
  }

  List<Provider> get services {
    if (widget.services == null) {
      return [];
    }
    return widget.services!.map((service) => service.provider).toList();
  }
}
