import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

import 'observable_model.dart';

/// A Gadget that represents a controller associated to a [ViewGadget].
///
/// It provides some useful shortcuts like getting a reference to
/// the [ObservableModel] or to some declared service.
///
/// To create a new controller, it's sufficient to extend this abstract class:
///
/// ```dart
/// class MyController extends ControllerGadget {
///   void onPressed() {
///     final value = model.getValue<MyValue>();
///     // Do some stuff with value...
///   }
/// }
/// ```
abstract class ControllerGadget {
  /// The [BuildContext] is automatically handled by the associated [ViewGadget].
  late BuildContext context;

  /// Gets a reference to the [ObservableModel].
  ObservableModel get model => context.read<ObservableModel>();

  /// Gets a reference to the requested service identified by type [T].
  T read<T>() => context.read<T>();

  ScaffoldMessengerState? get scaffold => ScaffoldMessenger.maybeOf(context);

  NavigatorState? get navigator => Navigator.maybeOf(context);

  ThemeData get theme => Theme.of(context);
}
