import 'dart:async';

import 'package:rxdart/rxdart.dart';

/// A function used to update the internal state of a value
/// stored in the [ObservableModel].
typedef Update<T> = T? Function(T value);

/// A function that converts a [Type] and an optional [String] into
/// a [String] used as a key for store values in the [ObservableModel].
///
/// ```dart
/// class CustomType { }
///
/// print(toValueKey(CustomType)); // CustomType_
/// print(toValueKey(CustomType, token: 'Token')); // CustomType_Token
/// ```
String toValueKey(Type type, {String? token}) {
  return '${type.toString()}_${token ?? ""}';
}

/// A Gadget used to store the global state of the application.
///
/// For each value provided to the [ObservableModel], a relative [GadgetSubject]
/// is created and stored in the internal `_subjectMap`.
/// This system represents the fundamental building block of the reactive
/// state management mechanism behind Flutter Gadgets. By using [GadgetSubject]s,
/// [SubscriberGadget]s can watch the generated streams and dynamically rebuild
/// whenever a new value is added or removed to one stream.
///
/// An [ObservableModel] instance is automatically created
/// and encapsulated in a [Service] by [AppGadget] during initialization.
class ObservableModel {
  final Map<String, GadgetSubject> _subjectMap = {};

  /// Retrieves a specific [GadgetSubject] from the map.
  ///
  /// If no subject is currently available, a new one will be
  /// automatically created.
  GadgetSubject<T?> getSubject<T>({String? token}) {
    GadgetSubject<T?>? subject =
        _subjectMap[toValueKey(T, token: token)] as GadgetSubject<T?>?;
    if (subject == null) {
      subject = _createSubject<T>();
      _subjectMap[toValueKey(T, token: token)] = subject;
    }
    return subject;
  }

  /// Retrieves the current value available in the relative [GadgetSubject]
  /// stored in the map.
  /// The map key is automatically generated using the [toValueKey] function.
  ///
  /// The returned value can be null.
  T? getValue<T>({String? token}) {
    // ignore: close_sinks
    final subject =
        _subjectMap[toValueKey(T, token: token)] as GadgetSubject<T?>?;
    if (subject == null) {
      return null;
    }
    return subject.valueOrNull;
  }

  /// Puts a value in the specified [GadgetSubject] stored in the map.
  ///
  /// If no [GadgetSubject] is associated to the key generated using
  /// the [toValueKey] function, a new one is automatically created.
  ///
  /// The provided value can be null.
  void putValue<T>({required T? value, String? token}) {
    // ignore: close_sinks
    GadgetSubject<T?>? subject =
        _subjectMap[toValueKey(T, token: token)] as GadgetSubject<T?>?;
    if (subject == null) {
      subject = _createSubject<T>();
      _subjectMap[toValueKey(T, token: token)] = subject;
    }
    subject.add(value);
  }

  /// Updates the internal state of a value using a given [Update] function.
  ///
  /// If the requested value is null, the provided [updateIfNull] value will be
  /// added to the [GadgetSubject].
  ///
  /// An exception will be thrown if the previous condition is true,
  /// but the [updateIfNull] value is null.
  T updateValue<T>({
    required Update<T> updateFunction,
    T? updateIfNull,
    String? token,
  }) {
    final value = getValue<T>(token: token);
    if (value == null && updateIfNull == null) {
      throw '''
      The requested value identified by ${toValueKey(T, token: token)} is null
      and no additional value has been provided using the updateIfNull
      parameter.
      ''';
    }
    if (value == null) {
      putValue<T>(value: updateIfNull, token: token);
      return updateIfNull!;
    }
    final updatedValue = updateFunction(value);
    putValue<T>(value: updatedValue, token: token);
    return value;
  }

  /// Puts a null value in the specified [GadgetSubject] stored in the map.
  ///
  /// If [closeSubject] is true, the selected [GadgetSubject] will be closed
  /// and removed from the map.
  ///
  /// If the specified [GadgetSubject] is null, the method will throw an exception.
  void removeValue<T>({bool closeSubject = false, String? token}) {
    // ignore: close_sinks
    GadgetSubject<T?>? subject =
        _subjectMap[toValueKey(T, token: token)] as GadgetSubject<T?>?;
    if (subject == null) {
      throw StateError('Tried to remove a value from an unexsisting subject.');
    }
    subject.add(null);
    if (closeSubject) {
      subject.close();
      _subjectMap.remove(toValueKey(T, token: token));
    }
  }

  /// Notifies the change of the internal state of a value
  /// stored in a [GadgetSubject].
  ///
  /// Is used to imperativly notify the change of the internal state
  /// of a specific value while its reference stays the same.
  ///
  /// Alternatively, the [updateValue] method can be used.
  void notifyFor<T>({String? token}) {
    // ignore: close_sinks
    final subject = getSubject<T>(token: token);
    subject.add(subject.valueOrNull);
  }

  /// Closes all the subjects stored in the map.
  void closeAllSubjects() {
    for (final subject in _subjectMap.values) {
      subject.close();
    }
  }

  /// Closes all the subjects and clears the _subjectMap.
  void dispose() {
    closeAllSubjects();
    _subjectMap.clear();
  }

  /// Creates a subject for a given type [T].
  GadgetSubject<T?> _createSubject<T>() => GadgetSubject<T?>();

  @override
  String toString() {
    final buffer = StringBuffer();
    for (final key in _subjectMap.keys) {
      buffer.write('$key - ${_subjectMap[key]?.valueOrNull}\n');
    }
    return buffer.toString();
  }
}

/// A custom rxdart [Subject] used internally by the [ObservableModel].
///
/// It's identical to a [PublishSubject], but it stores a reference to
/// the last value added to the stream.
class GadgetSubject<T> extends Subject<T> {
  GadgetSubject._(StreamController<T> controller, Stream<T> stream)
      : super(controller, stream);

  T? _lastValue;

  factory GadgetSubject(
      {void Function()? onListen,
      void Function()? onCancel,
      bool sync = false}) {
    // ignore: close_sinks
    final controller = StreamController<T>.broadcast(
      onListen: onListen,
      onCancel: onCancel,
      sync: sync,
    );

    return GadgetSubject<T>._(
      controller,
      controller.stream,
    );
  }

  @override
  void onAdd(T event) => _lastValue = event;

  T? get valueOrNull => _lastValue;
}
