import 'dart:async';

import 'package:collection/collection.dart';
import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';

import 'observable_model.dart';

/// A function used to build a Widget given a value retrieved
/// from the [ObservableModel].
typedef SubscriberGadgetBuilder<T> = Widget Function(
  BuildContext context,
  T value,
);

/// A function used to build a Widget when a given value - retrieved
/// from the [ObservableModel] - is null.
typedef NullGadgetBuilder = Widget Function(BuildContext context);

/// A function used to select the properties of a given value in order to
/// determine if the [SubscriberGadget] needs to be rebuilt.
typedef PropertiesSelector<T> = List<Object> Function(T value);

/// A Gadget used to observe a specified value from the [ObservableModel].
///
/// Using the type [T] and the optional [token], a [SubscriberGadget] registers
/// itself as a listener for the subject associated to [T] and [token] in the
/// [ObservableModel] and rebuilds itself each time a new value is added to the
/// subject - either by using [ObservableModel.putValue], [ObservableModel.updateValue]
/// or [ObservableModel.notifyFor].
///
/// When [propertiesSelector] is specified, [builder] will be called only if the current
/// values of the selected properties and the previous ones - compared using
/// [DeepCollectionEquality] from the package `collection` -  are different.
///
/// If the [useEquality] flag is true, whenever a rebuild is requested, the [builder]
/// function will be called only if the current value of the observed object and the
/// previous one - compared using the `==` operator - are different.
///
/// If the current value of the observed object is null, [SubscriberGadget] will use
/// the [nullBuilder] function instead of the [builder].
/// An exception will be thrown if the previous condition is true, but the [nullBuilder]
/// value is null.
///
/// It is used as follows:
///
/// ```dart
/// SubscriberGadget<MyValue>(
///   builder: (context, value) => Text('My value $value'),
///   nullBuilder: (context) => Text('My value is null...'),
/// )
/// ```
///
/// ```dart
/// SubscriberGadget<MyValue>(
///   propertiesSelector: (value) => [value.firstProperty]
///   builder: (context, value) => Text('First property $value.firstProperty'),
///   nullBuilder: (context) => Text('My value is null...'),
/// )
/// ```
///
/// ```dart
/// SubscriberGadget<int>(
///   useEquality: true,
///   builder: (context, value) => Text('int value $value'),
///   nullBuilder: (context) => Text('int value is null...'),
/// )
/// ```
class SubscriberGadget<T> extends StatefulWidget {
  const SubscriberGadget(
      {Key? key,
      required this.builder,
      this.nullBuilder,
      this.propertiesSelector,
      this.token,
      this.useEquality = false})
      : super(key: key);

  final SubscriberGadgetBuilder<T> builder;
  final NullGadgetBuilder? nullBuilder;
  final PropertiesSelector<T>? propertiesSelector;
  final bool useEquality;

  final String? token;

  @override
  _SubscriberGadgetState<T> createState() => _SubscriberGadgetState<T>();
}

class _SubscriberGadgetState<T> extends State<SubscriberGadget<T>> {
  late StreamSubscription _subscription;
  T? _currentValue;
  List<Object?>? _currentPropertiesState;

  @override
  void initState() {
    super.initState();
    final model = context.read<ObservableModel>();
    // ignore: close_sinks
    final subject = model.getSubject<T>(token: widget.token);
    _currentValue = subject.valueOrNull;
    if (_currentValue != null && widget.propertiesSelector != null) {
      _currentPropertiesState = widget.propertiesSelector!(_currentValue!);
    }
    _subscription = subject.listen((value) => _update(value));
  }

  @override
  void dispose() {
    _subscription.cancel();
    super.dispose();
  }

  void _update(T? value) {
    const equality = DeepCollectionEquality();
    if (value != null && widget.propertiesSelector != null) {
      if (_currentPropertiesState == null ||
          !equality.equals(
            _currentPropertiesState,
            widget.propertiesSelector!(value),
          )) {
        setState(
          () => _currentPropertiesState = widget.propertiesSelector!(value),
        );
      }
      return;
    }
    if (!widget.useEquality || _currentValue != value) {
      setState(() => _currentValue = value);
    }
  }

  @override
  Widget build(BuildContext context) {
    if (_currentValue == null && widget.nullBuilder == null) {
      throw '''
      The current value retrieved from the ObservableModel is null, but
      no nullBuilder has been specified for the SubscriberGadget.
      ''';
    }
    if (_currentValue == null) {
      return widget.nullBuilder!(context);
    }
    return widget.builder(context, _currentValue!);
  }
}
