import 'package:flutter/widgets.dart';

import 'controller_gadget.dart';

/// A function used to initialize a given [ControllerGadget].
typedef InitController<T extends ControllerGadget> = void Function(
  T controller,
);

/// A function used to dispose a given [ControllerGadget].
typedef DisposeController<T extends ControllerGadget> = void Function(
  T controller,
);

/// A function used to build a Widget given a [ControllerGadget].
typedef ViewGadgetBuilder<T extends ControllerGadget> = Widget Function(
  BuildContext context,
  T controller,
);

/// A Gadget used to create a portion of the UI while having access to a [ControllerGadget].
///
/// By using a [ViewGadget], it is possible to clearly decouple the creation of the UI - handled
/// by the [builder] function - and the management of UI events - handled by the callback
/// functions in the [ControllerGadget] - similarly to the Model-View-Controller architecture.
class ViewGadget<T extends ControllerGadget> extends StatefulWidget {
  const ViewGadget({
    Key? key,
    this.init,
    required this.builder,
    this.dispose,
    required this.controller,
  }) : super(key: key);

  final InitController<T>? init;
  final DisposeController<T>? dispose;
  final ViewGadgetBuilder<T> builder;
  final T controller;

  @override
  _ViewGadgetState<T> createState() => _ViewGadgetState<T>();
}

class _ViewGadgetState<T extends ControllerGadget>
    extends State<ViewGadget<T>> {
  @override
  void initState() {
    super.initState();
    if (widget.init != null) {
      widget.init!(widget.controller);
    }
  }

  @override
  void dispose() {
    if (widget.dispose != null) {
      widget.dispose!(widget.controller);
    }
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    widget.controller.context = context;
    return widget.builder(context, widget.controller);
  }
}
