import 'package:flutter/widgets.dart';
import 'package:provider/provider.dart';

import 'observable_model.dart';
import 'subscriber_gadget.dart';

/// A function used to initialize the internal state of
/// a given a view-model object [T].
typedef InitViewModel<T> = void Function(T viewModel);

/// A function used to dispose the internal state of
/// a given a view-model object [T].
typedef DisposeViewModel<T> = void Function(T viewModel);

/// A function used to build a Widget given a view-model object [T].
typedef ViewModelGadgetBuilder<T> = Widget Function(
  BuildContext context,
  T viewModel,
);

/// A Gadget that encapsulates and handles the state of a view model.
///
/// During initialization, the view-model object [T] is automatically registered
/// into the [ObservableModel] and the optional [init] callback is executed.
/// In this way, eventual changes to the view-model object can trigger the
/// [builder] function, effectively updating the UI.
///
/// When the widget is disposed, the view-model object [T] is automatically
/// removed from the [ObservableModel] and the optional [dispose] callback is executed.
/// The [dispose] callback can be used to manually dispose stateful components of
/// the given view-model object (like a [TextEditingController]).
class ViewModelGadget<T> extends StatefulWidget {
  const ViewModelGadget({
    Key? key,
    this.init,
    required this.builder,
    this.dispose,
    required this.viewModel,
  }) : super(key: key);

  final InitViewModel<T>? init;
  final DisposeViewModel<T>? dispose;
  final ViewModelGadgetBuilder<T> builder;
  final T viewModel;

  @override
  _ViewModelGadgetState<T> createState() => _ViewModelGadgetState<T>();
}

class _ViewModelGadgetState<T> extends State<ViewModelGadget<T>> {
  late ObservableModel model;

  @override
  void initState() {
    super.initState();
    if (widget.init != null) {
      widget.init!(widget.viewModel);
    }
    model = context.read<ObservableModel>()
      ..putValue<T>(value: widget.viewModel);
  }

  @override
  void dispose() {
    if (widget.dispose != null) {
      widget.dispose!(model.getValue<T>()!);
    }
    model.removeValue<T>(closeSubject: true);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return SubscriberGadget<T>(
      builder: (context, viewModel) => widget.builder(context, viewModel),
    );
  }
}
